"""
String matching automaton helpers.

"""
from ahocorasick import Automaton
from unidecode import unidecode


def normalize(string):
    """
    Normalize string for purposes of indexing/querying string matching automaton.

    """
    return unidecode(string).lower().strip()


def build_automaton(vocabulary, normalize=normalize):
    """
    Build a aho-corasick string matching automaton from vocabulary,
    given as an iterable over words to create dictionary from.

    """
    automaton = Automaton()
    for word in vocabulary:
        automaton.add_word(normalize(word), (len(word),))
    automaton.make_automaton()

    return automaton


def boundary_check(string, start_idx, end_idx):
    """
    Check that given character indexes into given string
    align with token boundaries.

    """
    start_clause = (
        (start_idx == 0) or
        string[start_idx-1] in (" ", ",")
    )
    end_clause = (
        (end_idx == len(string)) or
        string[end_idx] in (" ", ",", ".")
    )

    return (start_clause and end_clause)



"""
Street address validators.

"""
import re
from abc import ABCMeta, abstractmethod

from pystreet.automaton import boundary_check, build_automaton, normalize
from pystreet.data import iter_country_names


NONSTRICT_REGEXP = re.compile(
    r"""
    (?:
        (?:\b[0-9]+\s+[a-z]+\W+)   # e.g. '123 foo', '456 bar st.'
    |
        (?:\b[a-z]+\s+[0-9]+[\-\W]+)   # e.g. '', 'foobar 123,' 'foobar 2-4, 111'
    )

    """,
    re.IGNORECASE | re.VERBOSE,
)


class AddressValidator(metaclass=ABCMeta):

    def __call__(self, string):
        return self.validate(string)

    @abstractmethod
    def validate(self, string):
        """
        Validate given string is a street address.

        """


class NonstrictAddressValidator(AddressValidator):

    def __init__(self, boundary_check=boundary_check, regexp=NONSTRICT_REGEXP):
        """
        Non-strict street address validation.
        This validator is applicable in the broadest scope of global street address
        where nothing further is known about the string to be validated.

        It uses a simple heuristic:

        * Try to find a country mention (by name / ISO alpha-2 or alpha-3 code)
          at word boundaries in string using a string matching automaton.
        * Try to find some sub-structure in string that vaguely "looks like an address"
          This is done using a regular expression which tries to find some combination
          of numeric and alphabetical tokens in string.

        """
        self.automaton = build_automaton(vocabulary=iter_country_names())
        self.boundary_check = boundary_check
        self.regexp = regexp

    def validate(self, string):
        normalized = normalize(string)
        if not self.regexp.search(normalized):
            # Try to minimally identify address-like substructure first
            raise ValueError(f"Could not identify an address-like substructure in string: {string}")

        for end_position, (length, *tail) in self.automaton.iter(normalized):
            end_idx = end_position + 1
            start_idx = end_idx - length

            if not self.boundary_check(string, start_idx, end_idx):
                # Make sure string match aligns with word boundaries
                continue

            # If we reached here, can short circuit and verify address
            return True

        raise ValueError(f"Could not find a country mention in string: {string}")


def nonstrict_address_validator():
    return NonstrictAddressValidator()



#!/usr/bin/env python
from setuptools import find_packages, setup


project = "pystreet"
version = "0.1.0"

setup(
    name=project,
    version=version,
    description="Street address parsing and normalization",
    author="Globality Engineering",
    author_email="engineering@globality.com",
    url="https://github.com/globality-corp/pystreet",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "pyahocorasick>=1.1.6",
        "pycountry>=17.9.23",
        "unidecode>=1.0.22",
    ],
    setup_requires=[
        "nose>=1.3.7",
    ],
    tests_require=[
        "coverage>=3.7.1",
        "parameterized>=0.6.1",
        "PyHamcrest>=1.9.0",
    ],
)



"""
Data helpers.

"""
import pycountry


def iter_country_names():
    for country in pycountry.countries:
        if hasattr(country, "official_name"):
            yield country.official_name
        yield country.name
        yield country.alpha_3
